import re

from django.core.exceptions import ValidationError


def validate_unsubscribe_link(value):
    if value is None or value == '':
        return

    if len(re.findall(r'\{\{\s*unsubscribe_link\s*\}\}', value)) <= 0:
        raise ValidationError('Should include "{{ unsubscribe_link }}"')



# Generated by Django 2.2.5 on 2019-09-11 10:23

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import email_notifications.validators
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Notification',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('subject', models.CharField(max_length=255)),
                ('message', models.TextField(validators=[email_notifications.validators.validate_unsubscribe_link])),
                ('html_message', models.TextField(blank=True, null=True, validators=[email_notifications.validators.validate_unsubscribe_link])),
                ('create_at', models.DateTimeField(auto_now_add=True)),
                ('status', models.IntegerField(choices=[(0, 'fail'), (1, 'undone'), (2, 'success')], default=1)),
                ('finish_at', models.DateTimeField(blank=True, null=True)),
                ('users', models.ManyToManyField(to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='Status',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.IntegerField(choices=[(0, 'subscribe'), (1, 'unsubscribe')], default=0)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, unique=True)),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'status',
                'verbose_name_plural': 'status',
                'unique_together': {('user', 'uuid')},
            },
        ),
    ]



import logging

from django.core.mail import EmailMultiAlternatives, get_connection
from django.conf import settings
from django.db.models import Q
from django.http.request import HttpRequest
from django.template import Template, Context
from django.urls import reverse_lazy


from .models import Notification, Status


logger = logging.getLogger('django')
from_email = getattr(settings, 'EMAIL_HOST_USER')
unsubscribe_url = getattr(settings, 'unsubscribe_url', 'unsubscribe')


def send_notification(request: HttpRequest, instance: Notification):
    users = list(
        instance.users
        .select_related('status')
        .filter(Q(status__status=Status.SUBSCRIBE) | Q(status__status=None))
        .exclude(Q(email__isnull=True) | Q(email__exact='')),
    )

    # Update user list, only save user that have been send.
    instance.users.clear()
    instance.users.add(*users)

    if not users:
        instance.success()
        return

    # Create status for no status users
    Status.objects.bulk_create([
        Status(user=user) for user in
        filter(lambda u: getattr(u, 'status', None) is None, users)
    ])

    messages = []
    for user in users:
        path = reverse_lazy(unsubscribe_url, kwargs={'uuid': user.status.uuid})
        context = Context({
            'user': user,
            'unsubscribe_link': request.build_absolute_uri(str(path)),
        })
        message = EmailMultiAlternatives(
            Template(instance.subject).render(context),
            Template(instance.message).render(context),
            from_email,
            [user.email],
        )
        if instance.html_message:
            message.attach_alternative(
                Template(instance.html_message).render(context),
                'text/html',
            )

        messages.append(message)

    with get_connection() as connection:
        try:
            connection.send_messages(messages)
        except Exception as e:
            error_message_template = 'Error while sending notification {}: {}'
            logger.error(error_message_template.format(instance.id, e))
            instance.fail()
        else:
            instance.success()



from django.apps import AppConfig


class EmailNotificationsConfig(AppConfig):
    name = 'email_notifications'
    verbose_name = 'Email Notifications'

    def ready(self):
        from . import signals  # noqa



from django.contrib import admin

from . import services
from .models import Status, Notification


@admin.register(Status)
class StatusAdmin(admin.ModelAdmin):
    list_display = ('user', 'status', 'uuid')
    list_filter = ('status',)
    readonly_fields = ('uuid',)

    def get_queryset(self, request):
        return super().get_queryset(request).select_related('user')


@admin.register(Notification)
class NotificationAdmin(admin.ModelAdmin):
    list_display = ('subject', 'status', 'create_at', 'finish_at')
    list_filter = ('status',)
    filter_horizontal = ('users',)
    add_fieldsets = (
        (None, {
            'fields': ('subject', 'message', 'html_message', 'users'),
        }),
    )

    def get_fieldsets(self, request, obj=None):
        if not obj:
            return self.add_fieldsets

        return super().get_fieldsets(request, obj)

    def get_readonly_fields(self, request, obj=None):
        if obj is not None:
            return (
                'finish_at', 'subject', 'message',
                'html_message', 'users', 'status', 'create_at',
            )

        return super().get_readonly_fields(request, obj)

    def save_related(self, request, form, formsets, change):
        super().save_related(request, form, formsets, change)
        if form.instance.status == Notification.UNDONE:
            services.send_notification(request, form.instance)



import uuid

from django.contrib.auth import get_user_model
from django.db import models
from django.utils.timezone import now

from .validators import validate_unsubscribe_link


User = get_user_model()


class Status(models.Model):
    SUBSCRIBE = 0
    UNSUBSCRIBE = 1
    STATUS_CHOICES = (
        (SUBSCRIBE, 'subscribe'),
        (UNSUBSCRIBE, 'unsubscribe'),
    )

    user = models.OneToOneField(User, models.CASCADE)
    status = models.IntegerField(choices=STATUS_CHOICES, default=SUBSCRIBE)
    uuid = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)

    class Meta:
        verbose_name = 'status'
        verbose_name_plural = 'status'
        unique_together = (
            ('user', 'uuid'),
        )

    def __str__(self):
        return 'User {} - {}'.format(self.user_id, self.get_status_display())

    def subscribe(self):
        self.status = self.SUBSCRIBE
        self.save()

    def unsubscribe(self, refresh_uuid=True):
        self.status = self.UNSUBSCRIBE
        if refresh_uuid:
            self.uuid = uuid.uuid4()

        self.save()


class Notification(models.Model):
    FAIL = 0
    UNDONE = 1
    SUCCESS = 2
    STATUS_CHOICES = (
        (FAIL, 'fail'),
        (UNDONE, 'undone'),
        (SUCCESS, 'success'),
    )

    subject = models.CharField(max_length=255)
    message = models.TextField(validators=[validate_unsubscribe_link])
    html_message = models.TextField(
        null=True,
        blank=True,
        validators=[validate_unsubscribe_link],
    )
    create_at = models.DateTimeField(auto_now_add=True)
    status = models.IntegerField(choices=STATUS_CHOICES, default=UNDONE)
    finish_at = models.DateTimeField(null=True, blank=True)
    users = models.ManyToManyField(User)

    def __str__(self):
        return self.subject

    def finish(self):
        self.finish_at = now()
        self.save()

    def success(self):
        self.status = self.SUCCESS
        self.finish()

    def fail(self):
        self.status = self.FAIL
        self.finish()

    def save(self, **kwargs):
        self.full_clean()  # Call full clean to run validators
        super().save(**kwargs)



import os

from setuptools import find_packages, setup


EXCLUDE_FROM_PACKAGES = [
    'core',
]


def read(filename):
    with open(os.path.join(os.path.dirname(__file__), filename)) as f:
        return f.read()


setup(
    name='ponddy-email-notification',
    version=os.environ['CIRCLE_TAG'],
    url='https://github.com/ponddy-edu/ponddy-email-notification',
    author='Arthur Chang',
    author_email='arthurc0102@gmail.com',
    description='Ponddy email notification package',
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=EXCLUDE_FROM_PACKAGES),
    include_package_data=False,
    install_requires=[
        'django',
    ],
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python :: 3',
    ],
    keywords='ponddy',
)



from django.views.generic import TemplateView
from django.views.generic.edit import FormView
from django.shortcuts import get_object_or_404
from django.urls import reverse_lazy

from .forms import UnsubscribeForm
from .models import Status


class UnsubscribeView(FormView):
    template_name = 'email-notifications/unsubscribe.html'
    form_class = UnsubscribeForm
    success_url = reverse_lazy('unsubscribe_done')

    _status = None

    def dispatch(self, request, *args, **kwargs):
        self._status = get_object_or_404(
            Status,
            uuid=self.kwargs['uuid'],
            status=Status.SUBSCRIBE,
        )
        return super().dispatch(request, *args, **kwargs)

    def form_valid(self, form):
        self._status.unsubscribe()
        return super().form_valid(form)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({'user': self._status.user})
        return context


class UnsubscribeDoneView(TemplateView):
    template_name = 'email-notifications/unsubscribe_done.html'



from django.contrib.auth import get_user_model
from django.db.models.signals import post_save
from django.dispatch import receiver

from .models import Status


User = get_user_model()


@receiver(post_save, sender=User)
def create_status(instance, created, **kwargs):
    if not created:
        return

    Status.objects.create(user=instance)
